// game mode mechanics

void checkLock() {
	// check lock-down
	if (isOccupied(currentX, currentY-1, currentMino, currentRotation)) {
		if (lockCountDown > 0)
				lockCountDown--;
		else
			hardDrop();
	}
}

void checkGravity() {
	// check gravity
	if (gravityTimer > 0) {
		gravityTimer--;
	} else {
		move(0, -1, false);
	}
}

void initializeGame() {

	currentX = 4;
	currentY = 20;

	timeElapsed = 0;
	totalClears = 0;
	totalPieces = 0;
	totalBombs = 0;
	totalTetrises = 0;
	totalTSpins = 0;
	totalCombos = 0;
	longestCombo = 0;
  combo = 0;
	score = 0;
	b2b = false;
	topOut = false;
	holdPiece = 9;
	attackName = "";

	// Initialize the playField variable
	for (int i=0; i<10; i++) {
		for (int j=0; j<40; j++) {
			playField[i][j] = 9;
		}
	}

	// Initialize the pieces sequence
	int[] tmp = {0, 1, 2, 3, 4, 5, 6, 0, 1, 2, 3, 4, 5, 6, 9};
	for (int i=0; i<15; i++) {
		pieceSequence[i] = tmp[i];
	}
	generateSequence();		
	for (int i=0; i<7; i++) {
		pieceSequence[i] = pieceSequence[i+7];
	}
	generateSequence();
	// shuffle piece sequence an extra time, for fun
	for (int i=0; i<7; i++) {
		pieceSequence[i] = pieceSequence[i+7];
	}
	generateSequence();
}

void playerInput() {

	// Hard Drop Key
	if (checkKey(hardDropKey)) {
		if (!hardDropPressed) {
			hardDrop();
		}
		hardDropPressed = true;
	} else
		hardDropPressed = false;

	// Hold Key
	if (checkKey(holdKey1) || checkKey(holdKey2)) {
		if (!holdPressed) {
			hold();
		}
		holdPressed = true;
	} else
		holdPressed = false;

	// Right Rotate Key
	if (checkKey(rightRotateKey1) || checkKey(rightRotateKey2)) {
		if (!rightRotatePressed) {
			rotate(1);
		}
		rightRotatePressed = true;
	} else
		rightRotatePressed = false;

	// Left Rotate Key
	if (checkKey(leftRotateKey1) || checkKey(leftRotateKey2)) {
		if (!leftRotatePressed) {
			rotate(3);
		}
		leftRotatePressed = true;
	} else
		leftRotatePressed = false;

	// 180 Rotate Key
	if (checkKey(flipRotateKey)) {
		if (!flipRotatePressed) {
			rotate(2);
		}
		flipRotatePressed = true;
	} else
		flipRotatePressed = false;
	
	// left and right keys move mino

	if (checkKey(moveLeftKey)) {
		if (delayShift[0] < setDAS) { // left was pressed and not disabled
			if (delayShift[0] == 0)
				move(-1, 0, zeroARR);
			else
				delayShift[0]--;
		} else if (delayShift[0] == setDAS) {	// left previously not pressed
			if (checkKey(moveRightKey)) {	// right was held down
				delayShift[1] = setDAS + 1;		// disable right
			}
			move(-1, 0, false);
			delayShift[0]--;
		}									// otherwise do nothing, disabled
	} else {									// left released/not pressed
		if (delayShift[1] > setDAS) { //right was disabled & pressed. enable it
			delayShift[1] = 8;
		}
		delayShift[0] = setDAS;		// reset left
	}

	if (checkKey(moveRightKey)) {
		if (delayShift[1] < setDAS) { // right was pressed and not disabled
			if (delayShift[1] == 0)
				move(1, 0, zeroARR);
			else
				delayShift[1]--;
		} else if (delayShift[1] == setDAS) {	// right previously not pressed
			if (checkKey(moveLeftKey)) {	// left was held down
				delayShift[0] = setDAS + 1;		// disable left
			}
			move(1, 0, false);
			delayShift[1]--;
		}									// otherwise do nothing, disabled
	} else {									// right released/not pressed
		if (delayShift[0] > setDAS) { //left was disabled & pressed. enable it
			delayShift[0] = 8;
		}
		delayShift[1] = setDAS;		// reset right
	}

	// soft drop
	if (checkKey(softDropKey)) 
		move(0, -1, sonicDrop);
}

void hardDrop() {

	lockCountDown = 60;
	delayLock = 6;
	gravityTimer = 60;
	justHeld = false;
	totalPieces++;

	// move all the way down
	move(0, -1, true);
	// for bomb mode, check for bomb clears
	if (mode == 4 || mode == 5) {
		for (int i=0; i<4; i++) {
			int checkPosX = currentX +
						tetromino[currentMino][currentRotation][i][0];
			int checkPosY = currentY +
						tetromino[currentMino][currentRotation][i][1] - 1;
			if (checkPosY >= 0) {
				if (playField[checkPosX][checkPosY] == 8) {
					totalBombs++;
					playField[checkPosX][checkPosY] = 7;
				}
			}
		}
	}

	// fill tetromino in the field
	for (int i=0; i<4; i++) {
		int fillPosX = currentX + tetromino[currentMino][currentRotation][i][0];
		int fillPosY = currentY + tetromino[currentMino][currentRotation][i][1];
		playField[fillPosX][fillPosY] = currentMino;
	}

	// check for line clears
	lineClear();

	// reset position and rotation, bring out next piece
	currentX = 4;
	currentY = 20;
	currentRotation = 0;
	nextPiece();
	if (isOccupied(4, 20, currentMino, 0)) {
		topOut = true;
	}
}

void rotate(int direction) {
	for (int i = 0; i<5; i++) {
		int kickX = currentX +
				kickOffset[currentMino][currentRotation][i][0] -
				kickOffset[currentMino][(currentRotation + direction)%4][i][0];
		int kickY = currentY +
				kickOffset[currentMino][currentRotation][i][1] -
				kickOffset[currentMino][(currentRotation + direction)%4][i][1];
		if ( !isOccupied(kickX, kickY, currentMino,
				(currentRotation+direction)%4)){
			currentX = kickX;
			currentY = kickY;
			currentRotation = (currentRotation + direction) % 4;

			if (i == 0)
				lastMoveRotate = 1;
			else lastMoveRotate = 2;

			break;
		}
	}
	if (lockCountDown < 50) {	// lenient lockdown
		if (delayLock > 0) {
			lockCountDown += 10;
			delayLock--;
		}
	}
}

void move(int moveX, int moveY, boolean allTheWay) {

	if (!isOccupied(currentX + moveX, currentY + moveY, currentMino,
		currentRotation)) {
			currentX += moveX;
			currentY += moveY;
			lastMoveRotate = 0;
			if (moveY == -1)
				gravityTimer = 60;
	}
	if (allTheWay) {
		while (!isOccupied(currentX + moveX, currentY + moveY, currentMino,
		currentRotation)) {
			currentX += moveX;
			currentY += moveY;
		}
	}
}

void hold() {
	if (!justHeld) {
		currentX = 4;
		currentY = 20;
		currentRotation = 0;
		if (holdPiece == 9) {
			holdPiece = currentMino;
			nextPiece();
		} else {
			int tmp = holdPiece;
			holdPiece = currentMino;
			currentMino = tmp;
		}
		justHeld = true;
	}
}

void lineClear() {
	boolean lineFilled;
	boolean wasBomb = false;
	int pcCheck = 0;
	int clearLineCount = 0;

	// t-spin detection
	boolean tSpin;
	int countCorners = 0;
	if (currentMino == 4 && lastMoveRotate > 0) {
		if (currentX + 1 < 0 || currentX + 1 > 9 || currentY + 1 < 0 ||
			currentY + 1 > 39 || playField[currentX+1][currentY+1] != 9)
			countCorners++;
		if (currentX + 1 < 0 || currentX + 1 > 9 || currentY - 1 < 0 ||
			currentY - 1 > 39 || playField[currentX+1][currentY-1] != 9)
			countCorners++;
		if (currentX - 1 < 0 || currentX - 1 > 9 || currentY + 1 < 0 ||
			currentY + 1 > 39 || playField[currentX-1][currentY+1] != 9)
			countCorners++;
		if (currentX - 1 < 0 || currentX - 1 > 9 || currentY - 1 < 0 ||
			currentY - 1 > 39 || playField[currentX-1][currentY-1] != 9)
			countCorners++;
	}
	if (countCorners >= 3)
		tSpin = true;
	else
		tSpin = false;

	// check for filled lines, top to bottom
	for (int j=39; j>=0; j--) {

		lineFilled = true;
		for (int i=0; i<10; i++){
			pcCheck += playField[i][j];
			if ((playField[i][j] > 7)) {	// has bomb or has hole
				lineFilled = false;
			}
		}

		if (lineFilled) {	// clear line and fill down
			if (playField[0][j] < 7) {		//not garbage/bomb line
				clearLineCount++;
				totalClears++;
			} else {
				wasBomb = true;		// is garbage/bomb line
			}
			for (int i=0; i<10; i++){
				pcCheck -= playField[i][j];
			}
			pcCheck += 90;
			for (int k = j; k<39; k++) {
				for (int l=0; l<10; l++) {
					playField[l][k] = playField[l][k+1];
				}
			}
		}
	}

	// score keeping
	if (clearLineCount > 0) {
		attackName = "";

		if (tSpin) {	// T-spin
			totalTSpins++;
			if (b2b)
				attackName = "B2B ";
			if (clearLineCount == 1 && lastMoveRotate == 2) {	// mini
				attackName = attackName + "T-spin mini";
				if (b2b)
					score += 2;
				else
					score += 1;
			} else {
				switch(clearLineCount) {
					case 1:
						attackName = attackName + "T-spin Single";
						break;
					case 2:
						attackName = attackName + "T-spin Double";
						break;
					default:
						attackName = attackName + "T-spin Triple";
						break;
				}
				if (b2b)
					score += 3*clearLineCount;
				else
					score += 2*clearLineCount;
			}
			b2b = true;

		} else {		// Double, Triple, Tetris
			if (clearLineCount == 4) {
				totalTetrises++;
				if (b2b)
					attackName = "B2B ";
				attackName = attackName + "Tetris";
				if (b2b)
					score += 6;
				else
					score += 4;
				b2b = true;
			} else {
				b2b = false;
				if (clearLineCount == 2)
					score += 1;
				else if (clearLineCount == 3)
					score += 2;
			}
		}

		if (pcCheck == 3600) {	// PC
			attackName = "Perfect Clear";
			score += 10;
		}

	}

	if ((clearLineCount > 0) || wasBomb) {	// combo count and scores
		combo++;
		if (combo > 1)
			totalCombos++;	//	don't sum combo = 1 (which isn't combo)

		if (combo > longestCombo)
			longestCombo = combo;
		if (((combo > 1) && (attackName == "") ||
				(match(attackName, "Combo") != null))
				|| (combo > 5))
			attackName = (combo-1) + " Combo";
		switch(combo) {
			case 0:	// shouldn't happen
				break;
			case 1: // no combo yet
				break;
			case 2: // 1 combo
				score += 1;
				break;
			case 3: // 2 combo
				score += 1;
				break;
			case 4: // 3 combo
				score += 2;
				break;
			case 5: // 4 combo
				score += 2;
				break;
			case 6: // 5 combo
				score += 3;
				break;
			case 7: // 6 combo
				score += 3;
				break;
			default:	// 7 and more combo
				score += 4;
				break;
		}
	} else
		combo = 0;
}

void generateSequence() {	// fills index 7~13 of piece sequence
	// takes the first half of existing sequence and scramble it
	// for the second half
	int[] tmpList = new int[7];
	for (int i=0; i<7; i++) {
		tmpList[i] = pieceSequence[i];
	}

	int tmpCount = 7;
	for (int i=0; i<7; i++) {
		int j = int(random(tmpCount));	// take a random element of the list
		pieceSequence[i+7] = tmpList[j];// and put in piece sequence
		for (;j<tmpCount-1; j++){			// take out the element
			tmpList[j] = tmpList[j+1];
		}
		tmpCount--;
	}
}

void nextPiece() {
	currentMino = pieceSequence[0];
	for (int i=0; i<14; i++) {
		pieceSequence[i] = pieceSequence[i+1];
	}
	if (pieceSequence[7] == 9)
		generateSequence();
}

boolean isOccupied (int posX, int posY, int minoType, int minoRotation) {

	for (int i=0; i<4; i++) {
		int checkX = posX + tetromino[minoType][minoRotation][i][0];
		int checkY = posY + tetromino[minoType][minoRotation][i][1];
		if (checkX < 0 || checkX > 9 || checkY < 0 || checkY > 39)
			return true;
		else if (playField[checkX][checkY] != 9)
			return true;
	}
	return false;
}

void bombRise(int pos) {
	
	// lift matrix up one row
	for (int i=0; i<10; i++) {
		for (int j=37; j>=0; j--) {
			playField[i][j+1] = playField[i][j];
		}
	}
	// add a garbage row
	for (int i=0; i<10; i++) {
		playField[i][0] = 7;
	}
	// add a bomb
	playField[pos][0] = 8;
	// if mino is on the bottom, lift that too
	if (isOccupied(currentX, currentY, currentMino, currentRotation)) {
		move(0, 1, false);
	}
}



